%The prognostics algorithm is the standard PSO-PF algorithm, which is implemented 
%in the MATLAB programming language for model-based prognostics
%Parameters to adjust for different situations:
      % M: the time(cycle) when the prediction starts
      % MM:the length of the data
      % Q1,Q2: the process noise covariance, which we should change in different case
      % R: measurement noise covariance,  which we should change in different case
      % X1,X2: the initial valueof the state,
      % N: the number of particles in the particle filter
      % c1,c2: two learning factor in PSO algorithm
      % MaxDT: the maximum iteration for PSO
      % w2: a parameter indicates the importance of the last vitesse, in the mfile, we provide different options of w2,
      % which we should adjust selon different situation.

clear all;clc;
load('F:\mobaohua\lab\Research\\B0005');% the data file path
steps1 = B0005.cycle; % save steps array to new variable
inds1 = []; % initialize index array
y=[];% capacity
for(i=1:length(steps1)) %select the capacity of the discharge cycle
    if(strcmp(steps1(i).type,'discharge')) 
        inds1 = [inds1, i];
        y(i)=steps1(i).data.Capacity;
        y(y==0)=[];
    end
end


T=1;%sample time
M=100;%the start of prediction
MM=length(y);%the length of the data
delta_w1=1e-4;
delta_w2=1e-4;
Q1=delta_w1*diag([1,1]);%process noise covariance
Q2=delta_w2*1;%process noise covariance
R=0.0001;%measurement noise covariance
F=[1,0;0,1];%transition matrix

X1=zeros(2,M);%%state 1
X2=zeros(1,M);%state 2
Z=zeros(1,M);%observation
randn('seed',1)
w1=sqrtm(Q1)*randn(2,M);
%w1=sqrtm(Q1)*(2*rand(2,M)-1);
randn('seed',2)
w2=sqrtm(Q2)*randn(1,M);
%w2=sqrtm(Q2)*(2*rand(1,M)-1);
randn('seed',3)
v=2*sqrtm(R)*randn(1,M);%process noise
%v=sqrtm(R)*(2*rand(1,M)-1);
X1(:,1)=[-0.2,6.5]';%the initialization
X2(:,1)=1.85;%the initialization
state0=X1(:,1);%the initial state

N=200;%number of particles in the particle filter
zPred=zeros(1,N);
Weight=zeros(1,N);
xparticle1=zeros(2,N);
xparticle2=zeros(1,N);
zparticle=zeros(1,N);

%%%% Initialize the particle filter.
for t=2:M
   X1(:,t)=F*X1(:,t-1)+w1(:,t);
   X2(:,t)=0.997*X2(:,t-1)+X1(1,t-1)*exp(-X1(2,t-1))+w2(:,t);
end
for t=1:M
    Z(1,t)=X2(1,t)+v(1,t);
end

for j=1:N  
    randn('seed',4)
    xparticle1(:,j)=X1(:,1)+sqrtm(Q1)*randn(2,1);
    randn('seed',5)
    xparticle2(:,j)=X2(:,1)+sqrtm(Q2)*randn(1,1);
end
Xpf1=zeros(2,M);
Xpf2=zeros(2,M);
xparticlePred1=zeros(2,N);
xparticlePred2=zeros(1,N);
Xpf(:,1)=state0;
start=M;
for t=2:start
    %sampling
    randn('seed',6);
    num1=randn(2,N);
    randn('seed',7);
    num2=randn(1,N);
    for k=1:N
        xparticlePred1(:,k)=F*xparticle1(:,k)+sqrtm(Q1)*num1(:,k);% Update state from t(k-1) to t(k)
        % Update state from t(k-1) to t(k)
        xparticlePred2(1,k)=0.9972*xparticle2(1,k)+xparticle1(1,k)*exp(-xparticle1(2,k))+sqrtm(Q2)*num2(1,k);
    end
    xparticlePred=[xparticlePred1;xparticlePred2];
 

 %%%%%%%%%%%%%%%PSO%%%%%%%%%%%%
   c1=0.0005;c2=0.0005;%learning factor
   %w2=(M-t)*M+0.5;
   MaxDT=55;%the maximum iteration
   D=3;%the dimension of thr state space
   threshold=1e-7;
   flag=true;
   vitess=zeros(D,N);% the vitess vector 
   randn('seed',8);
for i=1:N
%     for j=1:D
%        vitess(j,i)=2*rand(1,1)-1; %the jth state of the ith particle
%     end
    vitess(:,i)=randn(D,1); %the jth state of the ith particle
end
   %%calculation of the particles' fiteness function , 
   %%initialization of pbest and gbest
   p=[];y2=[];
for i=1:N
    p(i)=Adapt2(xparticlePred(:,i),y(1,t));
    y2(:,i)=xparticlePred(:,i);%pbest of each particle
end
 pg=xparticlePred(:,1);% initialization of gbest
for i=2:N
    if Adapt2(xparticlePred(:,i),y(1,t))>Adapt2(pg,y(1,t));
        pg=xparticlePred(:,i);%the update of gbest
    end
end
CurCount=1;
%%the main recursion
Pbest=1;
for CurCount=1:MaxDT
      w2=0.02; %different choice of w2
      % w2=(MaxDT-CurCount)/MaxDT+0.4;% minW=0.5,maxW=1.5
      %w2 = 0.4 + (0.9-0.4) * exp(-3*(CurCount/MaxDT).^2);
      %w2=0.4*(0.9/0.4)^(1/(1+10*CurCount/MaxDT));
     %w2=(1.5-0.5)*(CurCount/MaxDT)^2+(0.5-1.5)*(2*CurCount/MaxDT)+1.2;
    for j=1:N
        % the update of the state 
        vitess(:,j)=w2*vitess(:,j)+c1*abs(randn)*(y2(:,j)-xparticlePred(:,j))+c2*abs(randn)*(pg-xparticlePred(:,j));
        xparticlePred(:,j)=xparticlePred(:,j)+vitess(:,j);
        if Adapt2(xparticlePred(:,j),y(1,t))>p(j)%the update of pbest
            p(j)=Adapt2(xparticlePred(:,j),y(1,t));
            y2(:,j)=xparticlePred(:,j);
        end
        if p(j)>Adapt2(pg,y(1,t)) %the update of gbest
            pg=y2(:,j);
        end
        
       if (abs(Adapt2(pg,y(1,t)))>threshold)
            flag=false;
       end
    end
    if (flag==true)
          break;
    end
   % Pbest=Adapt(pg,y(1,t),t);
end
    xparticlePred1(1,:)=xparticlePred(1,:);
    xparticlePred1(2,:)=xparticlePred(2,:);
    xparticlePred2(1,:)=xparticlePred(3,:);
    
    %calculation of the importance weight after PSO
    for k=1:N
        zPred(1,k)=xparticlePred2(1,k);%the prediction of the observation
        z1=y(1,t)-zPred(1,k);
        Weight(1,k)=exp(-(z1)^2/2/R);%the importance weight
    end
    %normalization of the weights
    Weight(1,:)=Weight(1,:)./sum(Weight(1,:));
    
    
    %%resampling
    index=sysresample(Weight(1,:));
    xparticle1(:,:)=xparticlePred1(:,index);% the new sample after resampling
    xparticle2(:,:)=xparticlePred2(:,index);% the new sample after resampling
    % the state after filtering
    Xpf1(:,t)=[mean(xparticle1(1,:)),mean(xparticle1(2,:))];  % The particle filter estimate is the mean of the particles.
    Xpf2(:,t)=mean(xparticle2(1,:));% The particle filter estimate is the mean of the particles.
    zparticle(1,:)=xparticle2(1,:);
    Zpf(1,t)=Xpf2(1,t);
    zpf(1,:)=zparticle(1,:);
end


start=M;
Zpf2(1,1)=Zpf(1,M);
Xf2(1,1)=start;
zpf2(:,1)=zpf(1,:)';
a=Xpf1(1,start);
b=Xpf1(2,start);
 for k=M+1:MM % prediction using the last updated parameter
     Xf2(1,k-start+1)=k;
     for i=1:N
        zpf2(i,k-start+1)=0.9972*zpf2(i,k-start)+a*exp(-b)+sqrt(R)*randn(1,1);% the update of the state of each particle
     end
    Zpf2(1,k-start+1)=mean(zpf2(:,k-start+1));
 end
 

EOL1=find(Zpf2<=1.4,1)+M-1;%%end of life

figure (1)
hold on;box on;
h1=plot(y(1,:),'-b');%measurement value
plot(Zpf(2:length(Zpf)),'-k','Linewidth',1.5);%estimation data using PSO-PF
h2=plot(Xf2,Zpf2,'-k','Linewidth',1.5);%prediction 
plot(128,1.2:0.01:1.9,'b.');%the EOL of battery B0005
plot(start,Zpf(1,M),'*');%the start prediction point
legend('PSO-PF');
h3=plot(0:180,linspace(1.4,1.4,181),'--m');%the threshold
h4=plot(linspace(start,start,71),1.2:0.01:1.9,'--c');
plot([EOL1,EOL1],[1.2,1.9],'-k','LineWidth',1.2)
axis([0 180 1.2 2]);
legend([h1(1),h2(1),h3(1),h4(1)],'real data','PSO-PF','threshold','start of prediction')

