%The prognostics algorithm is the standard QPSO-PF algorithm, which is implemented 
%in the MATLAB programming language for model-based prognostics
%Parameters to adjust for different situations:
      % M: the time(cycle) when the prediction starts
      % MM:the length of the data
      % Q1,Q2: the process noise covariance, which we should change in different case
      % R: measurement noise covariance,  which we should change in different case
      % X1,X2: the initial valueof the state, 
      % N: the number of particles in the particle filter
      % MaxDT: the maximum iteration for QPSO
      % b: the only parameter to be adjusted selon different situation


clear all;clc;
load('F:\mobaohua\lab\Research\\B0005');% the data file path
steps1 = B0005.cycle; % save steps array to new variable
inds1 = []; % initialize index array
y=[];% capacity
for(i=1:length(steps1)) %select the capacity of the discharge cycle
    if(strcmp(steps1(i).type,'discharge')) 
        inds1 = [inds1, i];
        y(i)=steps1(i).data.Capacity;
        y(y==0)=[];
    end
end

T=1;%sample time
M=70;%the start of prediction
MM=170;%the length of the data
delta_w1=1e-4;
Q1=delta_w1*diag([1,1]);%process noise covariance
Q2=delta_w1*1;%process noise covariance
R=0.0001;%process noise covariance
F=[1,0;0,1];%transition matrix

X1=zeros(2,M);%state 1
X2=zeros(1,M);%state 2
Z=zeros(1,M);%observation
randn('seed',1)
w1=sqrtm(Q1)*randn(2,M);
randn('seed',2)
w2=sqrtm(Q2)*randn(1,M);
randn('seed',3)
v=2*sqrtm(R)*randn(1,M);%process noise
X1(:,1)=[-0.2,6.5]';%the initialization %-0.3 6.5
X2(:,1)=1.85;%the initialization
state0=X1(:,1);%the initial state
N=200;%number of particles in the particle filter
zPred=zeros(1,N);
Weight=zeros(1,N);
xparticle1=zeros(2,N);
xparticle2=zeros(1,N);
zparticle=zeros(1,N);
%%%% Initialize the particle filter.
for t=2:M
   X1(:,t)=F*X1(:,t-1)+sqrtm(Q1)*w1(:,t);
   X2(:,t)=0.997*X2(:,t-1)+X1(1,t-1)*exp(-X1(2,t-1))+sqrtm(Q2)*w2(:,t);
end
for t=1:M
    Z(1,t)=X2(1,t)+sqrtm(R)*v(1,t);
end


for j=1:N   %Initialize the particle filter.
    randn('seed',4)
    xparticle1(:,j)=X1(:,1)+sqrtm(Q1)*randn(2,1);
    randn('seed',5)
    xparticle2(:,j)=X2(:,1)+sqrtm(Q2)*randn(1,1);
end
Xpf1=zeros(2,M);
Xpf2=zeros(2,M);
xparticlePred1=zeros(2,N);
xparticlePred2=zeros(1,N);
xparticlePred=zeros(3,N);
pbest=zeros(3,N);
Xpf(:,1)=state0;
start=M;
ErrorQPSOPF=0;
for t=2:start
    %sampling
    for k=1:N
        xparticlePred1(:,k)=F*xparticle1(:,k)+sqrtm(Q1)*randn(2,1);% Update state from t(k-1) to t(k)
        xparticlePred2(1,k)=0.9972*xparticle2(1,k)+xparticle1(1,k)*exp(-xparticle1(2,k))+sqrtm(Q2)*randn(1,1);% Update state from t(k-1) to t(k)
    end
    xparticlePred=[xparticlePred1;xparticlePred2];
 
    %calculation of the importance weight
    for k=1:N
        zPred(1,k)=xparticlePred2(1,k);%the prediction of the observation
        z1=y(1,t)-zPred(1,k);
        Weight(1,k)=exp(-(z1)^2/2/R);%the importance weight
    end
     
 %%%%%%%%%%%%%%%QPSO%%%%%%%%%%%%
  MaxDT=15;%the maximum iteration
  popsize=N;%number of the particles
  g=0;
  g=find(Weight==max(Weight));%find the particle with the highest weight
  gbest=xparticlePred(:,g); % gbest
  for i=1:N
      pbest(:,i)=xparticlePred(:,i);
  end
  
  threshold=1e-6;
  for i=1:MaxDT
      %b=0.2;  % different choice of b
      b=0.9-(i)/(MaxDT)*0.5;
      %b=0.9-(i-1)/(MaxDT-1)*0.2;
      %b=0.4*(0.9/0.4)^(1/(1+10*i/MaxDT));
      %b=0.5*(1/0.5)^(1/(1+10*i/MaxDT));
      % b=0.5 + (1-0.5) * exp(-3*(i/MaxDT).^2);
      %b = 0.4 + (0.9-0.4) * exp(-3*(i/MaxDT).^2);
      %w2=0.4*(0.9/0.4)^(1/(1+10*CurCount/MaxDT));
      %b=(1-0.5)*(MaxDT-i)/MaxDT+0.5;
      pbestTrans=pbest';
      mbest=(sum(pbestTrans)/popsize)';
      flag=true;
      
     for j=1:popsize
         % the update formule
          epsiln=1/i*abs(Adapt2(mbest,y(1,t))-Adapt2(pbest,y(1,t)));              
          c=unifrnd(0,1);
          p(:,1)=c*pbest(:,j)+(1-c)*pbest(:,g);
          u=unifrnd(0.01,1);
          if (unifrnd(0,1)>=0.5)
              xparticlePred(:,j)=p(:,1)-b*abs(mbest-xparticlePred(:,j))*log(1/(u+0.001));
          else
              xparticlePred(:,j)=p(:,1)+b*abs(mbest-xparticlePred(:,j))*log(1/(u+0.001));
          end
          
          
          if(Adapt2(xparticlePred(:,j),y(1,t))>Adapt2(pbest(:,j),y(1,t)))%the update of pbest
                  pbest(:,j)=xparticlePred(:,j);
          end
          
          if (Adapt2(pbest(:,j),y(1,t))>Adapt2(gbest,y(1,t)))%the update of gbest
            g=j;
            gbest=pbest(:,j);
          end
%       end
        if (abs(Adapt2(pbest(:,j),y(1,t)))>threshold)
            flag=false;
        end
   end
      if (flag==true)
          break;
      end
 end
  
    xparticlePred1(1,:)=xparticlePred(1,:);
    xparticlePred1(2,:)=xparticlePred(2,:);
    xparticlePred2(1,:)=xparticlePred(3,:);
    
    %calculation of the importance weight after QPSO
    for k=1:N
        zPred(1,k)=xparticlePred2(1,k);%the prediction of the observation
        z1=y(1,t)-zPred(1,k);
        Weight(1,k)=exp(-(z1)^2/2/R);%the importance weight
    end
    %normalization of the weights
    Weight(1,:)=Weight(1,:)./sum(Weight(1,:));
    
    
    %%resampling
    index=sysresample(Weight(1,:));
    xparticle1(:,:)=xparticlePred1(:,index);% the new sample after resampling
    xparticle2(:,:)=xparticlePred2(:,index);% the new sample after resampling
     % the state after filtering
    Xpf1(:,t)=[mean(xparticle1(1,:)),mean(xparticle1(2,:))];% The particle filter estimate is the mean of the particles.
    Xpf2(:,t)=mean(xparticle2(1,:));% The particle filter estimate is the mean of the particles.
    zparticle(1,:)=xparticle2(1,:);
    Zpf(1,t)=Xpf2(1,t);
    zpf(1,:)=zparticle(1,:);
end

start=M;
Zpf2(1,1)=Zpf(1,M);
Xf2(1,1)=start;
zpf2(:,1)=zpf(1,:)';
a=Xpf1(1,start);
b=Xpf1(2,start);
 for k=M+1:MM % prediction using the last updated parameter
     Xf2(1,k-start+1)=k;
     for i=1:N
        zpf2(i,k-start+1)=0.9972*zpf2(i,k-start)+a*exp(-b)+sqrt(R)*randn(1,1);% the update of the state of each particle
     end
    Zpf2(1,k-start+1)=mean(zpf2(:,k-start+1));
 end
 
EOL1=find(Zpf2<=1.4,1)+M-1;%%end of life

figure (1)
hold on;box on;
h1=plot(y(1,:),'-b','Linewidth',1.5);%measurement value
h2=plot(Xf2,Zpf2,'-r','Linewidth',1.5);%estimation data using QPSO-PF
plot(Zpf(2:length(Zpf)),'-r','Linewidth',1.5);%prediction 
plot(128,1.2:0.01:1.9,'b.');%the EOL of battery B0005
plot(start,Zpf(1,M),'*');%the start prediction point
h3=plot(0:180,linspace(1.4,1.4,181),'--m');%the threshold
h4=plot(linspace(start,start,71),1.2:0.01:1.9,'--c');
plot([EOL1,EOL1],[1.2,1.9],'-r','LineWidth',1.2)
axis([0 180 1.2 2]);
legend([h1(1),h2(1),h3(1),h4(1)],'real data','QPSO-PF','threshold','start of prediction')


